# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines types for `CheckpointableHandler`."""

from typing import Any, Awaitable, Protocol, Type, TypeVar, runtime_checkable
from orbax.checkpoint.experimental.v1._src.path import types as path_types


T = TypeVar('T')
AbstractT = TypeVar('AbstractT')


@runtime_checkable
class StatefulCheckpointable(Protocol[T]):
  """An interface that defines save/load logic for a `checkpointable` object."""

  async def save(
      self, directory: path_types.PathAwaitingCreation
  ) -> Awaitable[None]:
    """Saves the given `checkpointable` to the given `directory`."""
    ...

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: T | None = None,
  ) -> Awaitable[None]:
    """Loads the checkpointable from the given `directory`."""
    ...


class CheckpointableHandler(Protocol[T, AbstractT]):
  """An interface that defines save/load logic for a `checkpointable` object.

  NOTE: Prefer to use `StatefulCheckpointable` interface when possible.

  A "checkpointable" is a fundamental concept in Orbax. A “checkpointable”
  refers to a logical piece of the checkpoint that is distinct in some way from
  other pieces. Checkpointables are separable; they may or may not be loaded
  concurrently and some may be omitted from the checkpoint entirely.
  Checkpointables are often represented by different types, and have different
  representations on disk. The quintessential example is model params vs.
  dataset.

  A PyTree of arrays, representing model parameters, is the most basic
  "checkpointable". A singular array is also a checkpointable.

  In most contexts, when dealing with just a PyTree, the API of choice is::

    ocp.save_pytree(directory, pytree)

  The concept of "checkpointable" is not so obvious in this case. When dealing
  with multiple objects, we can use::

    ocp.save_checkpointables(
        directory,
        dict(
            pytree=model_params,
            dataset=dataset_iterator,
            # other checkpointables, e.g. extra metadata, etc.
        ),
    )

  Now, it is easy to simply skip loading the dataset, as is commonly desired
  when running evals or inference::

    ocp.load_checkpointables(
        directory,
        dict(
            pytree=abstract_model_params,
        ),
    )
    # Equivalently,
    ocp.load_pytree(directory, abstract_model_params)

  With the methods defined in this Protocol (`save`, `load`),
  logic within the method itself is executed in the main thread,
  in a blocking fashion. Additional logic can be executed in the background by
  returning an `Awaitable` function (which itself may return a result).

  Let's look at some suggestions on how to implement a `CheckpointableHandler`.
  TODO(b/398249409): Include more details on implementing this Protocol.

  First, take a look at
  orbax/checkpoint/experimental/v1/_src/testing/handler_utils.py
  for some toy implementations used for unit testing.

  Here are some details on how to implement `is_handleable` and
  `is_abstract_handleable`.

  For example, if a handler may be defined as follows::

    class FooHandler(CheckpointableHandler[Foo, AbstractFoo]):

      def is_handleable(self, checkpointable: Foo) -> bool:
        return isinstance(foo, Foo)

      def is_abstract_handleable(
          self, abstract_checkpointable: AbstractFoo) -> bool:
        return isinstance(abstract_foo, AbstractFoo)

  This is simple because the handler only works with `Foo` and `AbstractFoo`.
  But the handler may work on more generic types. In a toy
  example, let's say we've developed an improved way of storing very large
  arrays, which is still suboptimal for more normal-sized arrays. We can
  implement the handler as::

    class FooHandler(CheckpointableHandler[jax.Array, jax.ShapeDtypeStruct]):

      def is_handleable(self, checkpointable: jax.Array) -> bool:
        return (
            isinstance(checkpointable, jax.Array)
            and checkpointable.size > LARGE_ARRAY_THRESHOLD
        )

      def is_abstract_handleable(
          self, abstract_checkpointable: jax.ShapeDtypeStruct) -> bool:
        return (
            isinstance(abstract_checkpointable, jax.ShapeDtypeStruct)
            and abstract_checkpointable.size > LARGE_ARRAY_THRESHOLD
        )

  In many cases, no information is needed for loading. In this case,
  `AbstractT` may be defined as `None`. For example::

    class FooHandler(CheckpointableHandler[Foo, None]):

      def is_handleable(self, checkpointable: Foo) -> bool:
        return isinstance(checkpointable, Foo)

      def is_abstract_handleable(self, abstract_checkpointable: None) -> bool:
        return abstract_checkpointable is None
  """

  async def save(
      self, directory: path_types.PathAwaitingCreation, checkpointable: T
  ) -> Awaitable[None]:
    """Saves the given `checkpointable` to the given `directory`.

    Save should perform any operations that need to block the main thread, such
    as device-to-host copying of on-device arrays. It then creates a background
    operation to continue writing the object to the storage location.

    IMPORTANT: Do not assume that `directory` already exists at the start of
    this method. All directories are created by upper layers of the Orbax
    library, for performance reasons in a multihost setting and because upper
    layers also need to modify the directories. Before engaging in any
    filesystem operations, wait for the directory to exist. For example::

      async def _background_save(
          self,
          directory: path_types.PathAwaitingCreation,
          checkpointable: T,
      ) -> None:
        directory = await directory.await_creation()
        # Write to `directory` here.
        ...

      async def save(
          self,
          directory: path_types.PathAwaitingCreation,
          checkpointable: T,
      ) -> Awaitable[None]:
        # OK to access path properties, as long as we don't touch the actual
        # directory in the filesystem.
        logging.info(directory.name)
        return self._background_save(directory, checkpointable)

    Args:
      directory: The directory to save the checkpoint to. Note that the
        directory should not be expected to exist yet - it is in the process of
        being created. To wait for it to be created, use `await_creation`,
        preferably in a background awaitable to avoid blocking the main thread.
      checkpointable: The checkpointable object to save.

    Returns:
      An `Awaitable`. This object represents the result of the save
      operation running in the background.
    """
    ...

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: AbstractT | None = None,
  ) -> Awaitable[T]:
    """Loads the checkpointable from the given `directory`.

    Args:
      directory: The directory to load the checkpoint from.
      abstract_checkpointable: An optional abstract representation of the
        checkpointable to load. If provided, this is used to provide properties
        to guide the restoration logic of the checkpoint. In the case of arrays,
        for example, this conveys properties like shape and dtype, for casting
        and reshaping. In some cases, no information is needed, and `AbstractT`
        may always be None. In other cases, the abstract representation may be a
        hard requirement for loading.

    Returns:
      An `Awaitable` that continues to load the checkpointable in the background
      and returns the loaded checkpointable when complete.
    """
    ...

  async def metadata(self, directory: path_types.Path) -> AbstractT:
    """Returns the metadata for the given `directory`.

    The logic in this method must be executed fully in the main thread; metadata
    access is expected to be cheap and fast.

    In many cases it is desirable to return additional metadata properties
    beyond the limited set in `AbstractT`. In this case, `AbstractT` should
    be subclasses, and this subclass can be returned from `metadata`.

    Args:
      directory: The directory where the checkpoint is located.

    Returns:
      AbstractT: The metadata is an `AbstractT`, which is the abstract
      representation of the checkpointable.
    """
    ...

  def is_handleable(self, checkpointable: Any) -> bool:
    """Returns whether the handler can handle the given checkpointable.

    The method should return `True` if it is possible to save such an object.

    See class docstring for more details.

    Args:
      checkpointable: Either a concrete checkpointable, for saving.

    Returns:
      True if the handler can handle the given checkpointable.
    """
    ...

  def is_abstract_handleable(self, abstract_checkpointable: Any) -> bool | None:
    """Returns whether the handler can handle the abstract checkpointable.

    The method should return `True` if it is possible to use the given
    `abstract_checkpointable` for loading a concrete `T`. Note that `None` is
    always considered handleable for loading, so this method does not need to
    check for it. If an implementation defines
    `AbstractT` as `None`, then this method should only return True for values
    of `None`.

    See class docstring for more details.

    Args:
      abstract_checkpointable: An abstract checkpointable, for loading.

    Returns:
      True if the handler can handle the given checkpointable.
      None if the handler cannot decide whether it can handle the abstract
      checkpointable and defers to the typestr.
    """
    ...


def typestr(handler_cls: Type[CheckpointableHandler]) -> str:
  """A name for the handler class that uniquely identifies it."""
  return f'{handler_cls.__module__}.{handler_cls.__qualname__}'



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Composite logging implementation for checkpointing."""

from typing import Any
from absl import logging

from orbax.checkpoint._src.logging import abstract_logger


def is_valid_logger(logger: abstract_logger.AbstractLogger):
  if not isinstance(logger, abstract_logger.AbstractLogger):
    logging.warning(
        'CompositeLogger only supports AbstractLogger instances.'
    )
    return False
  if isinstance(logger, CompositeLogger):
    logging.warning(
        'CompositeLogger doesn\'t support instance of itself.'
    )
    return False
  return True


class CompositeLogger(abstract_logger.AbstractLogger):
  """Composite logger implementation."""

  def __init__(
      self,
      *loggers: abstract_logger.AbstractLogger,
  ):
    """CompositeLogger constructor.

    Args:
      *loggers: List of loggers to be used.
    """
    self._known_loggers: list[abstract_logger.AbstractLogger] = []
    for logger in loggers:
      if is_valid_logger(logger):
        self._known_loggers.append(logger)

  def log_entry(self, entry: Any):
    """Logs an informational message.

    Args:
        entry: Additional structured data to be included in the log record.
    """
    for logger in self._known_loggers:
      logger.log_entry(entry)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for Orbax.

NOTE: Functions in this file are deprecated in favor of corresponding functions
in sub-modules. Please use those functions instead, and do not add new
functions here.

TODO(b/266449081) Increase unit test coverage.
"""

from typing import Any

from etils import epath
import jax
import numpy as np
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import async_utils
from orbax.checkpoint._src.path import step as step_lib
from orbax.checkpoint._src.tree import utils as tree_utils


TMP_DIR_SUFFIX = step_lib.TMP_DIR_SUFFIX
# TODO(b/260759189): Deprecate this prefix when no longer in use by JAX MG.
_AGGREGATED_PREFIX = 'AGGREGATED://'
# Used in a msgpack checkpoint file to denote a leaf value that has been written
# individually. Typically, this may indicate an array that was written using
# Tensorstore rather than its value being directly stored in the msgpack file.
# To avoid duplication, we replace the value with a placeholder prefix and other
# relevant information (see functions below).
_PLACEHOLDER_PREFIX = 'PLACEHOLDER://'
PyTree = Any

sync_global_processes = multihost.sync_global_processes
sync_global_devices = multihost.sync_global_processes
broadcast_one_to_all = multihost.broadcast_one_to_all
reached_preemption = multihost.reached_preemption
is_primary_host = multihost.is_primary_host


async_makedirs = async_utils.async_makedirs
async_write_bytes = async_utils.async_write_bytes
async_exists = async_utils.async_exists

is_gcs_path = step_lib.is_gcs_path
checkpoint_steps = step_lib.checkpoint_steps
any_checkpoint_step = step_lib.any_checkpoint_step
is_checkpoint_finalized = step_lib.is_checkpoint_finalized
is_tmp_checkpoint = step_lib.is_tmp_checkpoint
tmp_checkpoints = step_lib.tmp_checkpoints
cleanup_tmp_directories = step_lib.cleanup_tmp_directories
get_save_directory = step_lib.get_save_directory
record_saved_duration = step_lib.record_saved_duration
step_from_checkpoint_name = step_lib.step_from_checkpoint_name
checkpoint_steps_paths = step_lib.checkpoint_steps_paths

deserialize_tree = tree_utils.deserialize_tree
from_flat_dict = tree_utils.from_flat_dict
# TODO: b/365169723 - Remove public access to this function.
from_flattened_with_keypath = tree_utils.from_flattened_with_keypath
serialize_tree = tree_utils.serialize_tree
to_flat_dict = tree_utils.to_flat_dict
is_sequence_key = tree_utils.is_sequence_key
is_dict_key = tree_utils.is_dict_key
tuple_path_from_keypath = tree_utils.tuple_path_from_keypath
get_key_name = tree_utils.get_key_name
is_empty_node = tree_utils.is_empty_node
is_empty_or_leaf = tree_utils.is_empty_or_leaf
to_shape_dtype_struct = tree_utils.to_shape_dtype_struct


def leaf_is_placeholder(leaf: Any) -> bool:
  """Determines if `leaf` represents a placeholder for a non-aggregated value."""
  return isinstance(leaf, str) and (
      leaf.startswith(_PLACEHOLDER_PREFIX)
      or leaf.startswith(_AGGREGATED_PREFIX)
  )


def leaf_placeholder(name: str) -> str:
  """Constructs value to act as placeholder for non-aggregated value."""
  return _PLACEHOLDER_PREFIX + name


def name_from_leaf_placeholder(placeholder: str) -> str:
  """Gets the param name from a placeholder with the correct prefix."""
  if not leaf_is_placeholder(placeholder):
    msg = (
        'Requested name from placeholder, but value did not contain required'
        ' prefix.'
    )
    raise ValueError(msg)
  if placeholder.startswith(_AGGREGATED_PREFIX):
    return placeholder.replace(_AGGREGATED_PREFIX, '', 1)
  elif placeholder.startswith(_PLACEHOLDER_PREFIX):
    return placeholder.replace(_PLACEHOLDER_PREFIX, '', 1)
  else:
    raise ValueError('Found placeholder beginning with unexpected prefix.')


def all_leaves_are_placeholders(tree: PyTree) -> bool:
  """Determines if all leaves in `tree` are placeholders."""
  return all(leaf_is_placeholder(leaf) for leaf in jax.tree.leaves(tree))


def pytree_structure(directory: epath.PathLike) -> PyTree:
  """Reconstruct state dict from saved model format in `directory`."""
  directory = epath.Path(directory)
  jax.monitoring.record_event('/jax/orbax/deprecation/inferred_structure')

  def add_nested_key(subtree, nested_key, key_name):
    if not nested_key:
      return subtree

    current = nested_key[0]

    if len(nested_key) == 1:
      assert current not in subtree
      subtree[current] = leaf_placeholder(key_name)
      return subtree

    subkeys = nested_key[1:]
    if current not in subtree:
      subtree[current] = {}
    subtree[current] = add_nested_key(subtree[current], subkeys, key_name)
    return subtree

  keys = directory.iterdir()
  tree = {}
  for k in keys:
    # Sharding file stores sharding data that is only used by orbax. Therefore,
    # it shouldn't be included here. See b/279969796 for more details.
    if k.name == '_sharding':
      continue
    if k.name == '_METADATA':
      continue
    # array_metadatas is not a checkpoint param. Only used when ocdbt is used.
    # ocdbt is still disabled in some projects like paxml.
    if k.name == 'array_metadatas':
      continue
    tree = add_nested_key(tree, k.name.split('.'), k.name)
  return tree


def is_scalar(x):
  return isinstance(x, (int, float, np.number))


def fully_replicated_host_local_array_to_global_array(
    arr: jax.Array,
) -> jax.Array:
  """Converts a host local array from to global jax.Array.

  In most cases, the local array is expected to have been produced by pmap.

  Args:
    arr: Host local array

  Returns:
    A global array.
  """
  # input `arr` is fully replicated, so it's shape is the global shape.
  global_shape = arr.addressable_data(0).shape

  # Create a 1D mesh to create fully replicated global jax.Array.
  mesh = jax.sharding.Mesh(np.array(jax.devices()), axis_names=('x',))
  partition_spec = (
      jax.sharding.PartitionSpec(None)
      if global_shape
      else jax.sharding.PartitionSpec()
  )
  # pmap-produced Array has a "scrambled" device order.
  dbs = sorted(
      [shard.data for shard in arr.addressable_shards],
      key=lambda x: list(x.devices())[0].id,
  )
  if jax.config.jax_pmap_no_rank_reduction and global_shape != arr.shape:
    dbs = [s[0] for s in dbs]
    global_shape = global_shape[1:]
  result = jax.make_array_from_single_device_arrays(
      global_shape, jax.sharding.NamedSharding(mesh, partition_spec), dbs
  )
  return result



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A signaling client interface and implementations."""

import functools
import threading
import time
from typing import Sequence
from absl import logging
import jax
from orbax.checkpoint._src.multihost import multihost
from typing_extensions import Protocol


class SignalingClient(Protocol):
  """Client that supports signaling between threads and processes."""

  def key_value_set(
      self, key: str, value: str, *, allow_overwrite: bool = False
  ):
    """Sets a key-value pair in the client.

    Args:
      key: The key to set.
      value: The value to set.
      allow_overwrite: Whether to allow overwriting an existing value for the
        given key.
    """

  def blocking_key_value_get(self, key: str, timeout_secs: int) -> str:
    """Gets the value for a given key in the client.

    Blocks until the key is set or the timeout is reached.

    Args:
      key: The key to get.
      timeout_secs: The timeout in seconds.
    """
    ...

  def key_value_try_get(self, key: str) -> str | None:
    """Tries to get the value for a given key in the client without blocking.

    Args:
      key: The key to get.
    """
    ...

  def key_value_delete(self, key: str):
    """Deletes the key-value.

    If the key is a directory, recursively clean up all key-values under the
    directory.

    Args:
      key: The key to delete.
    """
    ...

  def key_value_dir_get(self, key: str) -> Sequence[tuple[str, str]]:
    """Gets all key-value pairs in the directory.

    Args:
      key: The key to get.
    """
    ...

  def wait_at_barrier(
      self,
      key: str,
      *,
      timeout_secs: int,
      process_ids: Sequence[int] | None = None,
  ):
    """Waits at a barrier identified by key.

    Args:
      key: The key to wait at.
      timeout_secs: The timeout in seconds.
      process_ids: The participating process ids.
    """
    ...


class JaxDistributedSignalingClient(SignalingClient):
  """A signaling client that uses a JAX distributed client.

  This class uses a JAX distributed client to implement the signaling client
  interface.
  """

  def __init__(self):
    super().__init__()
    self._client = multihost.get_jax_distributed_client()

  def key_value_set(self, key: str, value: str, allow_overwrite: bool = False):
    """Sets a key-value pair in the client.

    Args:
      key: The key to set.
      value: The value to set.
      allow_overwrite: Whether to allow overwriting an existing value for the
        given key.

    Raises:
      JaxRuntimeError: If key value set fails.
    """
    self._client.key_value_set(key, value, allow_overwrite=allow_overwrite)

  def blocking_key_value_get(self, key: str, timeout_secs: int) -> str:
    """Gets the value for a given key in the client.

    Blocks until the key is set or the timeout is reached.

    Args:
      key: The key to get.
      timeout_secs: The timeout in seconds.

    Returns:
      The value associated with the key.

    Raises:
      JaxRuntimeError: If the key value get fails.
    """
    return str(self._client.blocking_key_value_get(key, timeout_secs * 1000))

  def key_value_try_get(self, key: str) -> str | None:
    """Tries to get the value for a given key in the client without blocking.

    Args:
      key: The key to get.

    Returns:
      The value associated with the key if it exists, otherwise None.
    """
    try:
      return str(self._client.key_value_try_get(key))
    except jax.errors.JaxRuntimeError:
      # Note that JaxRuntimeError may represent issues other than key not found,
      # but we are catching all such issues and returning None.
      logging.vlog(
          1,
          "JaxRuntimeError raised while trying to get key '%s'.",
          key,
          exc_info=True,
      )
      return None

  def key_value_delete(self, key: str):
    """Deletes the key-value.

    If the key is a directory, recursively clean up all key-values under the
    directory.

    Args:
      key: The key to delete.
    """
    self._client.key_value_delete(key)

  def key_value_dir_get(self, key: str) -> Sequence[tuple[str, str]]:
    """Gets all key-value pairs in the directory.

    Args:
      key: The key to get.

    Returns:
      A sequence of (key, value) tuples.
    """
    if not key.endswith("/"):
      raise ValueError(f"Key '{key}' must end with a slash.")
    return self._client.key_value_dir_get(key)

  def wait_at_barrier(
      self,
      key: str,
      *,
      timeout_secs: int,
      process_ids: Sequence[int] | None = None,
  ):
    """Waits at a barrier identified by key.

    Args:
      key: The key to wait at.
      timeout_secs: The timeout in seconds.
      process_ids: The participating process ids.

    Raises:
      TimeoutError: If the timeout is reached.
    """
    try:
      self._client.wait_at_barrier(
          key, timeout_secs * 1000, process_ids=process_ids
      )
    except jax.errors.JaxRuntimeError as e:
      # Note that JaxRuntimeError may represent issues other than timeouts, but
      # we aim to catch all such issues in caller layers.
      error_msg = f"Timeout waiting at barrier '{key}'."
      if process_ids is not None:
        error_msg += f" Barrier used process ids: {process_ids}."
      error_msg += f" The original error raised was: {str(e)}"
      raise TimeoutError(error_msg) from e


class ThreadSafeKeyValueSignalingClient(SignalingClient):
  """A thread-safe key-value store supporting basic operations and blocking get.

  This should only be used in a single controller setup as it does not support
  interaction with other processes.

  This class uses a dictionary internally to implement signaling client
  interface and protects concurrent access using a threading.Lock. It also uses
  a threading.Condition variable to allow threads to wait efficiently for a key
  to be set.
  """

  def __init__(self):
    super().__init__()
    self._data: dict[str, str] = {}
    self._lock = threading.Lock()
    self._condition = threading.Condition(self._lock)

  def key_value_set(
      self, key: str, value: str, allow_overwrite: bool = False
  ) -> None:
    """Sets the value for a given key in a thread-safe manner.

    If the key already exists and allow_overwrite is True, its value is
    overwritten, else an error is raised.
    Notifies any threads waiting on this key via the condition variable.

    Args:
        key: The key to set. Must be hashable.
        value: The value to associate with the key.
        allow_overwrite: Whether to allow overwriting an existing value for the
          given key.

    Raises:
        KeyError: If the key already exists and allow_overwrite is False.
    """
    with self._lock:
      if key in self._data and not allow_overwrite:
        raise KeyError(f"Key '{key}' already exists.")
      self._data[key] = value
      # Notify potentially waiting threads that the state has changed
      # (specifically, a key has been added or updated).
      self._condition.notify_all()

  def blocking_key_value_get(self, key: str, timeout_secs: int) -> str:
    """Gets the value for a key.

    Blocks until the key exists or a timeout occurs.

    Args:
        key: The key whose value is to be retrieved.
        timeout_secs: The maximum time in seconds to wait for the key to be set.

    Returns:
        The value associated with the key.

    Raises:
        TimeoutError: If the key is not set within the specified timeout.
    """
    end = time.time() + timeout_secs
    with self._condition:  # Acquires the underlying lock
      while key not in self._data:
        notified_in_time = self._condition.wait(timeout=end - time.time())

        if not notified_in_time:
          # Double-check after timeout before raising error
          if key not in self._data:
            raise TimeoutError(f"Timeout waiting for key '{key}'")

      return self._data[key]

  def key_value_try_get(self, key: str) -> str | None:
    """Tries to get the value for a key without blocking.

    Args:
        key: The key whose value is to be retrieved.

    Returns:
        The value associated with the key if it exists, otherwise None.
    """
    with self._lock:
      return self._data.get(key)

  def key_value_delete(self, key: str):
    """Deletes the key-value.

    If the key is a directory - ends with '/', recursively clean up all
    key-values under the directory.

    Args:
        key: The key to delete.
    """
    is_directory_key = key.endswith("/")
    with self._lock:
      if is_directory_key:
        keys_to_delete = [
            full_key
            for full_key in self._data
            if key == full_key or full_key.startswith(key)
        ]
        for k in keys_to_delete:
          del self._data[k]
      else:
        if key in self._data:
          del self._data[key]

  def key_value_dir_get(self, key: str) -> Sequence[tuple[str, str]]:
    """Gets all key-value pairs in the directory.

    Args:
      key: The key to get.

    Returns:
      A sequence of (key, value) tuples.
    """
    if not key.endswith("/"):
      raise ValueError(f"Key '{key}' must end with a slash.")
    keys_to_fetch = [
        full_key
        for full_key in self._data
        if key == full_key or full_key.startswith(key)
    ]
    return [(k, self._data[k]) for k in keys_to_fetch]

  def wait_at_barrier(
      self,
      key: str,
      *,
      timeout_secs: int,
      process_ids: Sequence[int] | None = None,
  ):
    """Waits at a barrier identified by key.

    Args:
      key: The key to wait at.
      timeout_secs: The timeout in seconds.
      process_ids: The participating process ids.
    """
    raise NotImplementedError(
        "`wait_at_barrier` is not implemented for"
        " `ThreadSafeKeyValueSignalingClient`."
    )


@functools.lru_cache()
def get_signaling_client() -> SignalingClient:
  """Returns the signaling client to use for the current environment."""
  if multihost.is_jax_distributed_client_initialized():
    logging.info("Using JaxDistributedSignalingClient")
    return JaxDistributedSignalingClient()
  else:
    process_count = multihost.process_count()
    if process_count > 1:
      raise RuntimeError(
          "ThreadSafeKeyValueSignalingClient should only be used in a single"
          f" controller setup, process count: {process_count}."
      )
    logging.info("Using ThreadSafeKeyValueSignalingClient")
    return ThreadSafeKeyValueSignalingClient()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test utils for work with PyTrees."""

from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost


def is_pytree_checkpoint_complete(directory):
  return (
      directory / format_utils.PYTREE_CHECKPOINTABLE_KEY / 'manifest.ocdbt'
  ).exists()


def get_d_files_mtimes(path: path_types.Path) -> list[int]:
  """Gets a list of last modified times for d files in a PyTree checkpoint.

  Assumes a structure like::

    path/
      <PYTREE_CHECKPOINTABLE_KEY>/
        ocdbt.process_0/
          d/
            <d_file_1>
            ...
            <d_file_n>
        ocdbt.process_1/
          ...

  The `path` and immediate subdirectories may optionally contain tmp suffixes.

  Args:
    path: The path to the checkpoint directory.

  Returns:
    A list of last modified times for d files in the checkpoint.
  """
  mtimes = []
  matching_dirs = list(path.parent.glob(f'{path.name}*'))
  if not matching_dirs:
    # Temp path not created yet.
    return []
  assert (
      len(matching_dirs) == 1
  ), f'Expected exactly one matching directory, got {matching_dirs}.'
  tmpdir = matching_dirs[0]
  matching_pytree_dirs = list(
      tmpdir.glob(f'{format_utils.PYTREE_CHECKPOINTABLE_KEY}*')
  )
  if not matching_pytree_dirs:
    # Temp path not created yet.
    return []
  assert len(matching_pytree_dirs) == 1, (
      'Expected exactly one matching pytree directory, got'
      f' {matching_pytree_dirs}.'
  )
  pytree_dir = matching_pytree_dirs[0]
  for idx in range(multihost.process_count()):
    d_path = pytree_dir / f'ocdbt.process_{idx}' / 'd'
    if not d_path.exists():
      continue
    for f in d_path.iterdir():
      # TensorStore tmp files can get deleted while iterating. We try to be
      # resilient to this.
      try:
        mtimes.append(f.stat().mtime)
      except OSError:
        pass
  return mtimes



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A class providing emergency checkpoint management.


WARNING: This class is experimental; do not use without specific approval.

NOTE: All classes within this module should be called across all *relevant*
processes. CheckpointManager is designed to be created and called across
*every* process. LocalCheckpointManager is designed to be created and called
across every process within *non-primary* slices. Similarly, a CheckpointManager
intended to work only with the persistent checkpoint on the primary slice should
always be called across all processes within the primary slice.
"""

import asyncio
import collections
import dataclasses
import enum
import functools
import itertools
import operator
import time
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Set

from absl import logging
from etils import epath
from etils import epy
import jax
import jax.numpy as jnp
import numpy as np
from orbax.checkpoint import abstract_checkpoint_manager
from orbax.checkpoint import args as args_lib
from orbax.checkpoint import checkpoint_manager
from orbax.checkpoint import checkpoint_utils
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.logging import abstract_logger
from orbax.checkpoint._src.logging import standard_logger
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.multihost import multislice
from orbax.checkpoint._src.path import step as step_lib
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint.experimental.emergency import local_checkpoint_data_debugging
from orbax.checkpoint.experimental.emergency import mesh_consistency
from orbax.checkpoint.experimental.emergency import process_metadata_checkpoint_handler
from typing_extensions import Self  # for Python version < 3.11


PyTree = checkpoint_manager.PyTree
CheckpointHandler = checkpoint_manager.CheckpointHandler
P = jax.sharding.PartitionSpec
PyTreeCheckpointHandler = pytree_checkpoint_handler.PyTreeCheckpointHandler
ProcessMetadataCheckpointHandler = (
    process_metadata_checkpoint_handler.ProcessMetadataCheckpointHandler
)
ChunkId = local_checkpoint_data_debugging.ChunkId
get_present_and_missing_chunks = (
    local_checkpoint_data_debugging.get_present_and_missing_chunks
)
RootMetadata = checkpoint_manager.RootMetadata
StepMetadata = checkpoint_manager.StepMetadata

_PRIMARY_REPLICA_ID = 0
_SECONDARY_REPLICA_ID = 1
_STATE_ITEM_NAME = 'state'
_PROCESS_METADATA_NAME = 'process_metadata'


local_all_steps_broadcast_counter = itertools.count()
find_complete_slice_broadcast_counter = itertools.count()


def _local_checkpoint_handler(
    multiprocessing_options: checkpoint_manager.MultiprocessingOptions,
) -> PyTreeCheckpointHandler:
  """Create a PyTreeCheckpointHandler for local checkpoints."""
  if multiprocessing_options.primary_host is not None:
    raise ValueError(
        'multiprocessing_options.primary_host must be set to None for local'
        ' checkpoints.'
    )
  local_registry = type_handlers.create_type_handler_registry(
      (
          jax.Array,
          type_handlers.ArrayHandler(
              primary_host=None,
              replica_id=None,
              use_replica_parallel=False,
          ),
      ),
  )
  return PyTreeCheckpointHandler(
      use_ocdbt=True,
      use_zarr3=True,
      multiprocessing_options=multiprocessing_options,
      type_handler_registry=local_registry,
  )


def _persistent_checkpoint_handler(
    multiprocessing_options: checkpoint_manager.MultiprocessingOptions,
) -> PyTreeCheckpointHandler:
  """Create a PyTreeCheckpointHandler for local checkpoints."""
  # TODO(b/372291557) Selection of replica_id=0 could be problematic if we can't
  # guarantee that the primary slice (in which the primary process is present)
  # always has the shard with shard.replica_id=0 for all available arrays.
  registry = type_handlers.create_type_handler_registry(
      (
          jax.Array,
          type_handlers.ArrayHandler(
              primary_host=multiprocessing_options.primary_host,
              replica_id=0,
              use_replica_parallel=False,
          ),
      ),
  )
  return PyTreeCheckpointHandler(
      use_ocdbt=True,
      use_zarr3=True,
      multiprocessing_options=multiprocessing_options,
      type_handler_registry=registry,
  )


@dataclasses.dataclass
class LocalCheckpointOptions:
  """Optional CheckpointManager arguments for saving local checkpoints.

  save_interval_steps:
    The interval at which checkpoints should be saved to local storage.
    Ensures checkpoints will only be saved every m steps. Defaults to 10.
  max_to_keep:
    Specifies the maximum number of local checkpoints to
    keep aside from the one currently being written. Older checkpoints are
    removed. When set, no more than (`max_to_keep` + 1) checkpoints will be
    present at any one time.
  read_only:
    If True, the local checkpoint manager will not save any checkpoints.
  should_save_fn:
    Predicate callable to check if given step can be saved. This callable
    accepts step number and optional latest step number as param and returns
    bool. If present then `save_interval_steps` and `save_on_steps` options are
    ignored.
  """

  save_interval_steps: int = 10
  max_to_keep: int = 1
  read_only: bool = False
  should_save_fn: Optional[Callable[[int, Optional[int]], bool]] = None

  debug_use_full_global_mesh: bool = False


@dataclasses.dataclass
class PersistentCheckpointOptions:
  """Optional CheckpointManager arguments for saving persistent checkpoints.

  save_interval_steps:
    The interval at which checkpoints should be saved to persistent storage.
    Ensures checkpoints will only be saved every n steps. Defaults to 1000.
  max_to_keep:
    If provided, specifies the maximum number of persistent checkpoints to
    keep. Older checkpoints are removed. By default, does not remove any old
    checkpoints. Must be None or non-negative. When set, checkpoints
    may be considered for deletion when there are more than `max_to_keep`
    checkpoints present.
  keep_period:
    If set, any existing checkpoints matching checkpoint_step % keep_period == 0
    will not be deleted.
  should_save_fn:
    Predicate callable to check if given step can be saved. This callable
    accepts step number and optional latest step number as param and returns
    bool. If present then `save_interval_steps` and `save_on_steps` options are
    ignored.
  """

  save_interval_steps: int = 1000
  max_to_keep: Optional[int] = None
  keep_period: Optional[int] = None
  should_save_fn: Optional[Callable[[int, Optional[int]], bool]] = None


@dataclasses.dataclass
class MultiprocessingOptions:
  """Options used to configure multiprocessing behavior.

  coordination_timeout_secs: The timeout in seconds for inter-process
    coordination. Essentially, this should represent the maximum amount of time
    that different processes can be "out of sync" by.
  """

  coordination_timeout_secs: int = 120


@dataclasses.dataclass
class CheckpointManagerOptions:
  """Optional arguments for CheckpointManager.

  local:
    Options relevant to the local checkpoints.
    See `LocalCheckpointOptions`.
  persistent:
    Options relevant to the persistent checkpoints. See
    `PersistentCheckpointOptions`.
  replica_axis_index:
    The index of the replica axis in the global mesh.
  step_name_format:
    NameFormat to build or find steps under input root directory. If provided,
    `step_prefix`, `step_format_fixed_length` are ignored.
  cleanup_tmp_directories:
    If True, cleans up any existing temporary directories
    on CheckpointManager creation.
  enable_async_checkpointing:
    Enable async saving.
  async_options:
    Used to configure properties of async behavior. See above.
  """

  local: LocalCheckpointOptions = dataclasses.field(
      default_factory=LocalCheckpointOptions
  )
  persistent: PersistentCheckpointOptions = dataclasses.field(
      default_factory=PersistentCheckpointOptions
  )

  replica_axis_index: int = 0

  step_name_format: step_lib.NameFormat[step_lib.Metadata] = (
      step_lib.standard_name_format()
  )
  cleanup_tmp_directories: bool = False
  enable_async_checkpointing: bool = True
  async_options: Optional[checkpoint_manager.AsyncOptions] = None
  multiprocessing_options: Optional[MultiprocessingOptions] = None


class _BarrierIdentifier(enum.Enum):
  """Identifies the barrier being run."""

  LOCAL_ALL_STEPS = 'local_all_steps'
  FIND_COMPLETE_SLICE = 'find_complete_slice'

  def get_counter(self) -> str:
    if self.name == self.LOCAL_ALL_STEPS.name:
      return str(next(local_all_steps_broadcast_counter))
    elif self.name == self.FIND_COMPLETE_SLICE.name:
      return str(next(find_complete_slice_broadcast_counter))
    else:
      raise ValueError(f'Unknown barrier identifier: {self.name}')


def _common_values_per_slice(
    per_process_values: Dict[int, Set[int]],
    global_mesh: jax.sharding.Mesh,
    *,
    replica_axis_index: int,
) -> Dict[int, Set[int]]:
  """Obtains values shared in common across all processes in each slice.

  Args:
    per_process_values: A mapping of process index to a list of values local to
      that process.
    global_mesh: The global mesh.
    replica_axis_index: The index of the replica axis in the global mesh.

  Returns:
    A mapping of slice index to a set of values shared in common across all
    processes in that slice. A value appearing in one process but not another
    in the same slice will not appear in the output.
  """
  total_num_slices = multislice.slice_count(
      global_mesh, replica_axis_index=replica_axis_index
  )
  num_processes_per_slice = (
      global_mesh.devices.size // total_num_slices // jax.local_device_count()
  )
  per_slice_values = collections.defaultdict(list)
  for pid, values in per_process_values.items():
    slice_id = multislice.process_slice_id(
        pid, global_mesh, replica_axis_index=replica_axis_index
    )
    per_slice_values[slice_id].extend(values)

  for slice_id, values in per_slice_values.items():
    counter = collections.Counter(values)
    common_values = [
        k for k in counter if counter[k] == num_processes_per_slice
    ]
    # Here `len(common_values)`` will be less than or equal to `len(values)`
    # because a value can only appear in `common_values` if it occurs
    # `num_processes_per_slice` times in `values`.
    if len(common_values) > len(values):
      raise AssertionError(
          f' len(common_values) ({common_values}) exceeded length of input'
          f' values ({values}).'
      )
    per_slice_values[slice_id] = common_values

  return {k: set(v) for k, v in per_slice_values.items()}


def _pad_steps(steps, target):
  return steps + [-1] * (target - len(steps))


def _process_local_to_global(
    values: Set[int],
    barrier_processes: Set[int],
    *,
    timeout: int,
    barrier_id: _BarrierIdentifier,
) -> Dict[int, Set[int]]:
  """Shares a sequence of host-local integers across given processes.

  Args:
    values: A set of local values. Each process has its own set of values.
    barrier_processes: A set of processes to share the set of values with.
    timeout: The timeout in seconds for inter-process coordination.
    barrier_id: Barrier identifier.

  Returns:
    A mapping of process index to the sequence of local values on that process.
    The result will have an entry for every process in `barrier_processes`.
  """
  barrier_name_and_id = f'{barrier_id.name}_{barrier_id.get_counter()}'
  client = multihost.get_jax_distributed_client()
  broadcast_dir_key = f'broadcast_{barrier_name_and_id}/'
  broadcast_dir_key = multihost._unique_barrier_key(broadcast_dir_key) + '/'  # pylint: disable=protected-access
  broadcast_key = broadcast_dir_key + str(multihost.process_index())
  client.key_value_set(broadcast_key, ','.join([str(s) for s in values]))

  barrier_key = f'barrier_{barrier_name_and_id}'
  barrier_key = multihost._unique_barrier_key(barrier_key)  # pylint: disable=protected-access
  logging.vlog(
      1,
      '[process=%s] Waiting at barrier %s',
      multihost.process_index(),
      barrier_key,
  )
  logging.vlog(
      1,
      '[process=%s] Barrier processes: %s',
      multihost.process_index(),
      barrier_processes,
  )
  client.wait_at_barrier(
      barrier_key,
      process_ids=list(barrier_processes),
      timeout_in_ms=timeout * 1000,
  )

  per_process_values = {
      int(k.split('/')[-1]): {int(s) for s in v.split(',')} if v else set()
      for k, v in client.key_value_dir_get(broadcast_dir_key)
  }
  assert set(per_process_values.keys()) == barrier_processes
  return per_process_values


def _all_devices_excepting_slice(
    devices: np.ndarray,
    *,
    replica_id: int = 0,
    replica_axis_index: int = 0,
) -> np.ndarray:
  return np.delete(devices, replica_id, axis=replica_axis_index)


def _get_global_broadcast_fn() -> Callable[[jax.Array], jax.Array]:
  """Returns a global broadcast function. More efficient to compile once."""
  # Compile function for global broadcast.
  slice_mesh = jax.sharding.Mesh(
      np.asarray(jax.devices()).reshape(
          multihost.process_count(), jax.local_device_count()
      ),
      ['host', 'dev'],
  )
  return jax.jit(
      lambda x: x,
      out_shardings=jax.sharding.NamedSharding(slice_mesh, P()),
  )


def _global_max(
    values: list[int], global_broadcast_fn: Callable[[jax.Array], jax.Array]
) -> list[int]:
  """Computes the global max of a list of values."""
  num_hosts = multihost.process_count()
  num_devices_per_host = jax.local_device_count()
  slice_mesh = jax.sharding.Mesh(
      np.asarray(jax.devices()).reshape(num_hosts, num_devices_per_host),
      ['host', 'dev'],
  )

  sdas = []
  for d in jax.local_devices():
    sdas.append(
        jax.device_put(np.asarray(values).reshape((1, 1, len(values))), d)
    )
  sharding = jax.sharding.NamedSharding(slice_mesh, P('host', 'dev'))
  # TODO(cpgaffney): Use jax.make_array_from_process_local_data.
  g_arr = jax.make_array_from_single_device_arrays(
      (num_hosts, num_devices_per_host, len(values)), sharding, sdas
  )
  result_arr = global_broadcast_fn(g_arr)
  result_arr = np.asarray(result_arr.addressable_data(0))

  # Checks that for every host, values are equal across local devices.
  assert (
      np.sum(result_arr, axis=1) / num_devices_per_host == result_arr[:, 0, :]
  ).all()
  # Select values from first device and compute max for each value across
  # hosts.
  return list(np.max(result_arr[:, 0, :], axis=0).astype(int))


class _LocalCheckpointManager(checkpoint_manager.CheckpointManager):
  """A checkpoint manager that checkpoints to local storage.

  Attributes:
    device_array: an ndarray representing all the devices running
      LocalCheckpointManager in the same global jax Mesh, importantly the first
      axis of the device_array is assumed to be the direction of device slices
      across which the Data Parallelism is happening.
  """

  # TODO: b/330585086 - Allow configuration of global mesh describing slices.
  # Validate against global meshes used for arrays in state.
  # TODO: b/330585086 - Support arbitrary items beyond state. We will have
  # to evaluate whether arbitrary items can be a good fit for local
  # checkpointing, given restore+broadcast requirements.
  def __init__(
      self,
      directory: epath.PathLike,
      global_mesh: jax.sharding.Mesh,
      *,
      primary_replica_id: int = _PRIMARY_REPLICA_ID,
      options: Optional[CheckpointManagerOptions] = None,
      metadata: Optional[dict[str, Any]] = None,
      logger: Optional[abstract_logger.AbstractLogger] = None,
  ):
    # TODO: b/330585086 - Fully support options.
    options = options or CheckpointManagerOptions()
    self._global_mesh = global_mesh
    self._replica_axis_index = options.replica_axis_index

    devices = np.asarray(self._global_mesh.devices)
    # Select all devices except those belonging to the primary replica.
    if not options.local.debug_use_full_global_mesh:
      devices = _all_devices_excepting_slice(
          devices,
          replica_id=primary_replica_id,
          replica_axis_index=self._replica_axis_index,
      )

    self._active_processes = multihost.unique_processes_from_devices(devices)
    multiprocessing_options = checkpoint_manager.MultiprocessingOptions(
        primary_host=None,
        active_processes=self._active_processes,
        barrier_sync_key_prefix='local',
    )
    async_options = options.async_options or checkpoint_manager.AsyncOptions()
    async_options.create_directories_asynchronously = False
    local_options = checkpoint_manager.CheckpointManagerOptions(
        save_interval_steps=options.local.save_interval_steps,
        max_to_keep=options.local.max_to_keep,
        step_name_format=options.step_name_format,
        should_save_fn=options.local.should_save_fn,
        create=False,
        # we always clean up local tmp directories explicitly
        cleanup_tmp_directories=False,
        async_options=async_options,
        multiprocessing_options=multiprocessing_options,
        enable_async_checkpointing=options.enable_async_checkpointing,
        read_only=options.local.read_only,
        single_host_load_and_broadcast=False,
        # enable_background_delete set to False to ensure gc is done before save
        enable_background_delete=False,
        save_root_metadata=False,
    )
    self._logger = logger or standard_logger.StandardLogger()
    self._coordination_timeout_secs = (
        options.multiprocessing_options or MultiprocessingOptions()
    ).coordination_timeout_secs
    super().__init__(
        directory,
        options=local_options,
        metadata=metadata,
        item_handlers=dict(
            state=_local_checkpoint_handler(multiprocessing_options),
            process_metadata=ProcessMetadataCheckpointHandler,
        ),
        logger=self._logger,
    )
    self._run_initial_garbage_collection()
    # Set additional properties.
    self._max_to_keep = options.local.max_to_keep
    self._local_options = options.local
    self._steps = list(self.all_steps(read=True))

  def _run_initial_garbage_collection(self):
    """Remove steps that might be left over from previous runs."""
    steps_to_remove = self._get_old_steps_to_remove()
    self._checkpoints.delete_if(lambda info: info.step in steps_to_remove)
    self._checkpoint_deleter.delete_steps(steps_to_remove)

  def local_host_steps(self, read: bool) -> Sequence[int]:
    """Returns steps known to local host."""
    # List of steps present in individual host storage.
    local_steps = list(super().all_steps(read))
    logging.info(
        'Found steps: %s in local host storage: %s.',
        local_steps,
        self.directory,
    )

    if len(local_steps) > self._max_to_keep:
      logging.error(
          'local_step on host %d exceeded `max_to_keep` %d',
          multihost.process_index(),
          self._max_to_keep,
      )

    return _pad_steps(local_steps, self._max_to_keep)

  def all_steps(self, read: bool = False) -> Sequence[int]:
    """Returns all steps tracked by the manager.

    Includes steps located in local storage.

    Args:
      read: If True, forces a read directly from the storage location.
        Otherwise, a cached result can be returned.

    Returns:
      A sequence of steps (integers)
    """
    if read:
      local_steps = set(self.local_host_steps(read))
      # Per-process mapping of the local steps that each process knows about.
      per_process_steps = _process_local_to_global(
          local_steps,
          self._active_processes,
          timeout=self._coordination_timeout_secs,
          barrier_id=_BarrierIdentifier.LOCAL_ALL_STEPS,
      )
      slice_id = multislice.process_slice_id(
          multihost.process_index(),
          self._global_mesh,
          replica_axis_index=self._replica_axis_index,
      )
      per_slice_steps = _common_values_per_slice(
          per_process_steps,
          self._global_mesh,
          replica_axis_index=self._replica_axis_index,
      )
      logging.info(
          'After broadcast, found steps %s shared between local slice'
          ' processes.',
          per_slice_steps[slice_id],
      )
      steps = functools.reduce(operator.ior, per_slice_steps.values(), set())
      self._steps = [x for x in steps if x != -1]
    return self._steps

  def latest_step(self) -> Optional[int]:
    """Returns the latest step saved in the local storage.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """
    return max(self._steps) if self._steps else None

  def save(
      self,
      step: int,
      args: Optional[args_lib.Composite] = None,
      metrics: Optional[PyTree] = None,
      force: Optional[bool] = False,
  ) -> bool:
    """Saves the checkpoint at the given step."""
    saved = super().save(step, args=args, metrics=metrics, force=force)
    if saved:
      self._steps.append(step)
      self._steps = self._steps[-self._max_to_keep :]
    return saved

  def reload(self):
    """Reloads internal properties.

    refreshes the cached list of globally available local checkpointed steps.
    """
    super().reload()
    self._steps = list(self.all_steps(read=True))


def _get_persistent_options(
    options: CheckpointManagerOptions,
    multiprocessing_options: checkpoint_manager.MultiprocessingOptions,
) -> checkpoint_manager.CheckpointManagerOptions:
  """Get options for persistent checkpoint manager."""
  return checkpoint_manager.CheckpointManagerOptions(
      save_interval_steps=options.persistent.save_interval_steps,
      max_to_keep=options.persistent.max_to_keep,
      keep_period=options.persistent.keep_period,
      step_name_format=options.step_name_format,
      create=False,
      cleanup_tmp_directories=options.cleanup_tmp_directories,
      async_options=options.async_options,
      multiprocessing_options=multiprocessing_options,
      enable_async_checkpointing=options.enable_async_checkpointing,
      should_save_fn=options.persistent.should_save_fn,
      save_root_metadata=False,
  )


class _MultisliceCheckpointManager(
    abstract_checkpoint_manager.AbstractCheckpointManager, epy.ContextManager
):
  """Provides both checkpoint management and emergency checkpointings.

  This class is an implementation layer for handling multislice checkpointing.

  This class composes a local and a persistent checkpoint managers. The local
  manager saves checkpoints frequently to a fast local storage (like RAMFS).
  When a complete checkpoint exists at least one slice, restoration is possible,
  and the slice broadcasts the checkpoint to others. Additionally, the
  persistent manager checkpoints less frequently to a remote file system (e.g.,
  GCS),
  providing a fail-safe if local checkpoints become unavailable due to issues
  like hardware failure or preemption.

  Usage::
    options = CheckpointManagerOptions(
        local=LocalCheckpointOptions(save_interval_steps=2, max_to_keep=2),
        persistent=PersistentCheckpointOptions(
            save_interval_steps=5, max_to_keep=3
        ),
        enable_async_checkpointing=use_async,
    )
    return _MultisliceCheckpointManager(
        local_directory=local_directory,
        persistent_directory=persistent_directory,
        global_mesh=global_mesh,
        abstract_state=abstract_state,
        options=options,
    )
  """

  # TODO: b/330585086 - Support arbitrary items beyond state. We will have
  # to evaluate whether arbitrary items can be a good fit for local
  # checkpointing, given restore+broadcast requirements.
  def __init__(
      self,
      local_directory: epath.PathLike,
      persistent_directory: epath.PathLike,
      global_mesh: jax.sharding.Mesh,
      abstract_state: PyTree,  # a single PyTree describing the state structure
      *,
      options: Optional[CheckpointManagerOptions] = None,
      metadata: Optional[dict[str, Any]] = None,
      logger: Optional[abstract_logger.AbstractLogger] = None,
  ):
    self._local_directory = epath.Path(local_directory)
    self._persistent_directory = epath.Path(persistent_directory)
    if not self._local_directory.exists():
      raise FileNotFoundError(
          f'Local directory {self._local_directory} must be created by the'
          ' caller.'
      )
    if not self._persistent_directory.exists():
      raise FileNotFoundError(
          f'Persistent directory {self._persistent_directory} must be created'
          ' by the caller.'
      )

    self._logger = logger or standard_logger.StandardLogger()
    # TODO: b/330585086 - Fully support options.
    options = options or CheckpointManagerOptions()
    self._replica_axis_index = options.replica_axis_index
    self._global_mesh = global_mesh
    logging.info(
        'Configured emergency.CheckpointManager with replica_axis_index=%d,'
        ' corresponding to "%s" in the global mesh.',
        self._replica_axis_index,
        self._global_mesh.axis_names[self._replica_axis_index],
    )

    self._abstract_state = abstract_state
    self._slice_id = multislice.process_slice_id(
        multihost.process_index(),
        self._global_mesh,
        replica_axis_index=self._replica_axis_index,
    )
    self._options = options
    self._metadata = metadata

    self._persistent_max_to_keep = self._options.persistent.max_to_keep
    self._local_max_to_keep = self._options.local.max_to_keep
    self._coordination_timeout_secs = (
        options.multiprocessing_options or MultiprocessingOptions()
    ).coordination_timeout_secs
    self._slice_count = multislice.slice_count(
        self._global_mesh, replica_axis_index=self._replica_axis_index
    )

    if len(global_mesh.devices.shape) <= self._replica_axis_index:
      raise AssertionError(
          f'replica_axis_index {self._replica_axis_index} is out of bound for'
          f' global_mesh.devices.shape {global_mesh.devices.shape}'
      )
    if self._slice_count <= 1:
      raise AssertionError(
          'To use this CheckpointManager, at least 2 data-parallel replicas are'
          ' needed.'
      )

    primary_replica_id = _PRIMARY_REPLICA_ID
    secondary_replica_id = _SECONDARY_REPLICA_ID

    self._persistent_primary_host = multislice.primary_process_in_slice(
        self._global_mesh,
        replica_axis_index=self._replica_axis_index,
        replica_id=primary_replica_id,
    )
    self._local_primary_host = multislice.primary_process_in_slice(
        self._global_mesh,
        replica_axis_index=self._replica_axis_index,
        replica_id=secondary_replica_id,
    )
    self._in_primary_slice = multislice.in_slice(
        multihost.process_index(),
        global_mesh,
        replica_axis_index=self._replica_axis_index,
        replica_id=primary_replica_id,
    )

    if self._in_primary_slice:
      persistent_multiprocessing_options = (
          checkpoint_manager.MultiprocessingOptions(
              primary_host=self._persistent_primary_host,
              active_processes=multihost.unique_processes_from_devices(
                  multislice.slice_devices(
                      self._global_mesh,
                      replica_axis_index=self._replica_axis_index,
                      replica_id=primary_replica_id,
                  )
              ),
              barrier_sync_key_prefix='persistent',
          )
      )
      self._persistent_checkpoint_manager = (
          self._make_persistent_checkpoint_manager(
              persistent_multiprocessing_options
          )
      )
    else:
      self._local_checkpoint_manager = self._make_local_checkpoint_manager(
          primary_replica_id
      )

    self._local_steps = []
    self._persistent_steps = []
    # clean up tmp directories in ram
    self._cleanup_local_tmp_directories()

    # Initialize step cache.
    self.all_steps(read=True)

    self._global_broadcast_fn = _get_global_broadcast_fn()

    logging.info(
        'Created emergency.CheckpointManager with slice_id=%d,'
        ' process_index=%d, jax.process_index=%d',
        self._slice_id,
        multihost.process_index(),
        jax.process_index(),
    )
    logging.vlog(1, 'Local devices: %s', jax.local_devices())

  def _cleanup_local_tmp_directories(self):
    logging.info(
        'Cleaning up existing temporary directories at %s.',
        self._local_directory,
    )
    tmp_files = step_lib.tmp_checkpoints(self._local_directory)
    for tmp_file in tmp_files:
      logging.info('Deleting temporary checkpoint: %s.', tmp_file)
      (self._local_directory / tmp_file).rmtree()

  def _make_persistent_checkpoint_manager(
      self,
      persistent_multiprocessing_options: checkpoint_manager.MultiprocessingOptions,
  ) -> checkpoint_manager.CheckpointManager:
    return checkpoint_manager.CheckpointManager(
        self._persistent_directory,
        options=_get_persistent_options(
            self._options, persistent_multiprocessing_options
        ),
        metadata=self._metadata,
        item_handlers=_persistent_checkpoint_handler(
            persistent_multiprocessing_options
        ),
        logger=self._logger,
    )

  def _make_local_checkpoint_manager(
      self, primary_replica_id: int = _PRIMARY_REPLICA_ID
  ) -> _LocalCheckpointManager:
    return _LocalCheckpointManager(
        self._local_directory,
        global_mesh=self._global_mesh,
        primary_replica_id=primary_replica_id,
        options=self._options,
        metadata=self._metadata,
        logger=self._logger,
    )

  @property
  def directory(self) -> epath.Path:
    return self._persistent_directory

  @property
  def in_primary_slice(self) -> bool:
    return self._in_primary_slice

  @property
  def global_mesh(self) -> jax.sharding.Mesh:
    return self._global_mesh

  def all_steps(self, read: bool = False) -> Sequence[int]:
    """Returns all steps tracked by the manager.

    Includes steps located in local as well as persistent storage.

    Args:
      read: If True, forces a read directly from the storage location.
        Otherwise, a cached result can be returned.

    Returns:
      A sequence of steps (integers)
    """
    logging.info('Retrieving all steps.')
    if read:
      per_slice_local_steps = self._get_per_slice_local_steps()
      self._local_steps = list(set.union(*per_slice_local_steps.values()))
      self._persistent_steps = step_lib.checkpoint_steps(
          self._persistent_directory
      )
    return list(set(self._local_steps) | set(self._persistent_steps))

  def latest_step(self) -> Optional[int]:
    """Returns the latest step saved.

    Includes steps located in local as well as persistent storage.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """
    logging.info('Retrieving latest step.')
    all_steps = self.all_steps()
    return max(all_steps) if all_steps else None

  def best_step(self) -> Optional[int]:
    """Returns the best step saved, as defined by `options.best_fn`.

    Includes steps located in local as well as persistent storage.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """
    raise NotImplementedError(
        'Metrics tracking not yet implemented for emergency.CheckpointManager.'
    )

  def reload(self):
    """Performs disk reads to ensure internal properties are up to date."""
    if self.in_primary_slice:
      self._persistent_checkpoint_manager.reload()
    else:
      self._local_checkpoint_manager.reload()

  def reached_preemption(self, step: int) -> bool:
    """Returns True if a preemption sync point has been reached."""
    return multihost.reached_preemption(step)

  def should_save(self, step: int) -> bool:
    """Returns True if a checkpoint should be saved for the current step.

    This depends the previous step and save interval.

    Args:
      step: int

    Returns:
      True if the checkpoint should be saved.
    """
    logging.info('Checking should_save at step: %d.', step)
    if self.in_primary_slice:
      should_save = self._persistent_checkpoint_manager.should_save(step)
    else:
      should_save = self._local_checkpoint_manager.should_save(step)
    return bool(_global_max([int(should_save)], self._global_broadcast_fn)[0])

  def delete(self, step: int):
    """Deletes a step checkpoint."""
    raise NotImplementedError(
        'Delete not yet implemented for emergency.CheckpointManager.'
    )

  def save(
      self,
      step: int,
      args: args_lib.Composite,
      *,
      force: bool = False,
  ) -> bool:
    """Returns True if a checkpoint was saved either locally or persistently."""
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'CheckpointManager:save_start',
            prefix='emergency_checkpoint_manager',
        ),
        record_event_name=(
            '/jax/orbax/write/checkpoint_start_sync_duration_secs'
        ),
    )

    if not isinstance(args, args_lib.Composite):
      raise ValueError(
          f'Expected args must be a Composite object, but got {type(args)}.'
      )

    if _STATE_ITEM_NAME not in args.keys():
      raise ValueError(
          f'{_STATE_ITEM_NAME} is a required key and should be'
          ' specified by the user.'
      )

    if _PROCESS_METADATA_NAME in args.keys():
      raise ValueError(
          f'{_PROCESS_METADATA_NAME} is a reserved key and should not be'
          ' specified by the user.'
      )

    # TODO: b/330608746 - implement save op on different slices
    persistent_saved = False
    local_saved = False
    if self.in_primary_slice:
      logging.info('Maybe saving at step %d (persistent).', step)
      persistent_saved = self._persistent_checkpoint_manager.save(
          step, args=args.state, force=force
      )
    else:
      logging.info('Maybe saving at step %d (local).', step)

      args_dict = dict(args.items())
      args_dict[_PROCESS_METADATA_NAME] = (
          process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
              global_mesh=self._global_mesh
          )
      )
      args = args_lib.Composite(**args_dict)

      local_saved = self._local_checkpoint_manager.save(
          step, args=args, force=force
      )

    start = time.time()
    saved = tuple(
        bool(e)
        for e in _global_max(
            [int(persistent_saved), int(local_saved)], self._global_broadcast_fn
        )
    )
    persistent_saved, local_saved = saved
    logging.info('Broadcast `saved` bool in %f seconds.', time.time() - start)

    if persistent_saved:
      self._persistent_steps.append(step)
      if self._persistent_max_to_keep is not None:
        self._persistent_steps = self._persistent_steps[
            -self._persistent_max_to_keep :
        ]
    if local_saved:
      self._local_steps.append(step)
      self._local_steps = self._local_steps[-self._local_max_to_keep :]

    return persistent_saved or local_saved

  def _get_per_slice_local_steps(self) -> Dict[int, Set[int]]:
    local_steps = set(step_lib.checkpoint_steps(self._local_directory))
    logging.info(
        'Found steps: %s in local host storage: %s.',
        local_steps,
        self._local_directory,
    )

    # The steps that each process actually has in local storage.
    per_process_steps = _process_local_to_global(
        local_steps,
        set(range(jax.process_count())),
        timeout=self._coordination_timeout_secs,
        barrier_id=_BarrierIdentifier.FIND_COMPLETE_SLICE,
    )
    logging.vlog(1, 'per_process_steps=%s', per_process_steps)
    per_slice_steps = _common_values_per_slice(
        per_process_steps,
        self._global_mesh,
        replica_axis_index=self._replica_axis_index,
    )
    logging.vlog(1, 'per_slice_steps=%s', per_slice_steps)
    return per_slice_steps

  def _find_slice_with_complete_local_checkpoint(self, step: int) -> int:
    """Return the slice id which has the step."""
    per_slice_steps = self._get_per_slice_local_steps()

    for slice_id, steps in per_slice_steps.items():
      if step in steps:
        return slice_id
    return -1

  def _get_single_slice_sharding(
      self,
      mesh: jax.sharding.Mesh,
      pspec: jax.sharding.PartitionSpec,
  ):
    """Get sharding for a single slice."""
    slice_devices = multislice.slice_devices(
        mesh,
        replica_id=self._slice_id,
        replica_axis_index=self._replica_axis_index,
    )
    single_slice_mesh_shape = [
        1 if i == self._replica_axis_index else d
        for i, d in enumerate(mesh.devices.shape)
    ]
    slice_mesh = jax.sharding.Mesh(
        slice_devices.reshape(single_slice_mesh_shape), mesh.axis_names
    )
    return jax.sharding.NamedSharding(slice_mesh, pspec)

  def _restore_from_local(
      self,
      step: int,
      restoring_slice_id: int,
      directory: Optional[epath.PathLike] = None,
  ) -> Any:
    logging.info(
        'emergency.CheckpointManager: restoring step=%s from local checkpoint'
        ' using slice_id: %s',
        step,
        restoring_slice_id,
    )
    step_stats = step_statistics.EmergencyRestoreStepStatistics()
    step_stats.checkpoint_manager_start_time = time.time()
    step_stats.step = step
    is_restoring_slice = restoring_slice_id == self._slice_id
    step_stats.is_restoring_slice = is_restoring_slice
    step_stats.in_primary_slice = self.in_primary_slice

    shape_dtypes, tree_defs = jax.tree.flatten(self._abstract_state)
    original_single_slice_shardings = jax.tree.map(
        lambda arr: self._get_single_slice_sharding(
            mesh=arr.sharding.mesh,
            pspec=arr.sharding.spec,
        ),
        self._abstract_state,
    )
    original_single_slice_shardings_tuple = tuple(
        jax.tree.flatten(original_single_slice_shardings)[0]
    )

    # Debug logging for sharding information.
    if logging.vlog_is_on(1):
      logging.vlog(
          1,
          'Debugging global restore_args based on abstract state. This uses the'
          ' user-provided mesh, and is not actually used for restoration.',
      )
      local_checkpoint_data_debugging.print_devices_indices_debug_info(
          checkpoint_utils.construct_restore_args(self._abstract_state)
      )
      logging.vlog(
          1,
          'Debugging single-slice restore_args based on abstract state. This'
          ' uses the user-provided mesh, and is not actually used for'
          ' restoration.',
      )
      local_checkpoint_data_debugging.print_devices_indices_debug_info(
          checkpoint_utils.construct_restore_args(
              self._abstract_state, original_single_slice_shardings
          )
      )

    if is_restoring_slice:
      logging.vlog(
          1, 'emergency.CheckpointManager: restoring from local checkpoint.'
      )
      restore_directory = self._options.step_name_format.find_step(
          epath.Path(directory or self._local_directory), step
      ).path
      step_stats.directory = str(restore_directory)
      (
          previous_distributed_to_device_ids,
          previous_device_ids,
      ) = ProcessMetadataCheckpointHandler().restore(
          restore_directory / _PROCESS_METADATA_NAME,
          process_metadata_checkpoint_handler.ProcessMetadataRestoreArgs(),
      )
      restore_mesh = mesh_consistency.consistent_restore_mesh_from_metadata(
          self._global_mesh,
          multihost.distributed_to_device_ids(),
          previous_distributed_to_device_ids=previous_distributed_to_device_ids,
          previous_device_ids=previous_device_ids,
      )
      restoring_processes = multihost.unique_processes_from_devices(
          multislice.slice_devices(
              restore_mesh,
              replica_id=self._slice_id,
              replica_axis_index=self._replica_axis_index,
          )
      )
      multiprocessing_options = checkpoint_manager.MultiprocessingOptions(
          primary_host=None,
          active_processes=restoring_processes,
          barrier_sync_key_prefix='local_restoring_slice',
      )
      local_state_handler = _local_checkpoint_handler(multiprocessing_options)

      restore_single_slice_shardings = jax.tree.map(
          lambda arr: self._get_single_slice_sharding(
              mesh=restore_mesh,
              pspec=arr.sharding.spec,
          ),
          self._abstract_state,
      )
      single_slice_restore_args = checkpoint_utils.construct_restore_args(
          self._abstract_state, restore_single_slice_shardings
      )

      # Directly use CheckpointHandler to restore. This is undesirable, but
      # allows us to avoid barrier issues that occur when calling
      # LocalCheckpointManager a different number of times on the non-primary
      # slices, which leads to
      # _module_unique_count getting out of sync.
      logging.vlog(
          1,
          'Restoring from %s',
          restore_directory / _STATE_ITEM_NAME,
      )
      if logging.vlog_is_on(1):
        logging.vlog(
            1,
            'Debugging single-slice restore_args used for restoration.',
        )
        asyncio.run(
            local_checkpoint_data_debugging.print_chunk_debug_info(
                restore_directory / _STATE_ITEM_NAME,
                single_slice_restore_args,
            )
        )
        local_checkpoint_data_debugging.print_devices_indices_debug_info(
            single_slice_restore_args
        )

      step_stats.checkpointer_start_time = time.time()
      args = args_lib.PyTreeRestore(
          item=self._abstract_state,
          restore_args=checkpoint_utils.construct_restore_args(
              self._abstract_state
          ),
      )
      single_slice_pytree = local_state_handler.restore(
          restore_directory / _STATE_ITEM_NAME,
          args=dataclasses.replace(
              args, restore_args=single_slice_restore_args
          ),
      )
      step_stats.checkpointer_duration_secs = (
          time.time() - step_stats.checkpointer_start_time
      )
      in_tree = tuple(jax.tree.flatten(single_slice_pytree)[0])
      if not np.array_equal(
          restore_mesh.device_ids, self._global_mesh.device_ids
      ):
        # User-provided mesh is usually optimized for performance.
        # But we permuted the original mesh so that we can read each locally
        # available shard correctly. This may cause performance issues.

        # Thus, we re-shard the array to follow the original mesh and layout.
        in_tree = mesh_consistency.consistent_restore_mesh_to_global_mesh(
            in_tree, original_single_slice_shardings_tuple
        )
    else:
      logging.vlog(
          1,
          'emergency.CheckpointManager: secondary slice, create zeros and'
          ' wait for broacast.',
      )

      @functools.partial(
          jax.jit,
          static_argnums=0,
          out_shardings=original_single_slice_shardings_tuple,
      )
      def create_zeros(shape_dtype_tup):
        return jax.tree.map(
            lambda sd: jnp.zeros(sd.shape, dtype=sd.dtype), shape_dtype_tup
        )

      zeros_pytree = create_zeros(tuple(shape_dtypes))
      in_tree = tuple(zeros_pytree)

    multihost.sync_global_processes('local_restore_pre_broadcast')

    start_broadcast = time.time()
    shared_states, _ = multislice.broadcast_one_replica_to_all(
        in_tree,
        self._global_mesh,
        replica_axis_index=self._replica_axis_index,
        is_source=is_restoring_slice,
    )
    broadcast_elapsed_s = time.time() - start_broadcast
    jax.monitoring.record_event_duration_secs(
        '/orbax/emergency/checkpoint/read/broadcast_duration_secs',
        broadcast_elapsed_s,
    )
    step_stats.broadcast_start_time = start_broadcast
    step_stats.broadcast_duration_secs = broadcast_elapsed_s
    step_stats.checkpoint_manager_duration_secs = (
        time.time() - step_stats.checkpoint_manager_start_time
    )
    self._logger.log_entry(dataclasses.asdict(step_stats))

    logging.info('Finished broadcasting in %.2f', broadcast_elapsed_s)

    return jax.tree.unflatten(tree_defs, shared_states)

  def _restore_from_persistent(
      self,
      step: int,
      directory: Optional[epath.PathLike] = None,
  ) -> Any:
    logging.info(
        'emergency.CheckpointManager: restoring step=%s from persistent'
        ' checkpoint in directory=%s',
        step,
        directory or self._persistent_directory,
    )
    args = args_lib.PyTreeRestore(
        item=self._abstract_state,
        restore_args=checkpoint_utils.construct_restore_args(
            self._abstract_state
        ),
    )

    # Create a temporarily read-only PersistentCheckpointManager that will
    # synchronize the restoration with global processes.
    persistent_options = checkpoint_manager.CheckpointManagerOptions(
        step_name_format=self._options.step_name_format,
        create=False,
        cleanup_tmp_directories=False,
        read_only=True,
        enable_async_checkpointing=False,
        multiprocessing_options=checkpoint_manager.MultiprocessingOptions(
            barrier_sync_key_prefix='persistent_global',
        ),
    )
    with checkpoint_manager.CheckpointManager(
        self._persistent_directory,
        options=persistent_options,
        metadata=self._metadata,
        item_handlers=PyTreeCheckpointHandler(
            use_ocdbt=True,
            use_zarr3=True,
        ),
    ) as pcm:
      try:
        return pcm.restore(step, args=args, directory=directory)
      except FileNotFoundError as e:
        raise FileNotFoundError(
            'No steps found in either local or persistent storage when'
            f' requesting restoration of step {step}.'
        ) from e

  def restore(
      self,
      step: Optional[int],
      args: args_lib.Composite | None = None,
  ) -> Any:
    del args
    if step is None:
      step = self.latest_step()
      if step is None:
        raise FileNotFoundError(
            'No steps found in persistent or local storage.'
        )
    logging.info('Restoring at step %d.', step)
    restoring_slice_id = self._find_slice_with_complete_local_checkpoint(step)
    if restoring_slice_id > -1:
      # restore from LCM
      return self._restore_from_local(
          step=step,
          restoring_slice_id=restoring_slice_id,
      )

    return self._restore_from_persistent(step=step)

  def item_metadata(self, step: int) -> Any:
    raise NotImplementedError(
        'Item metadata not yet implemented for emergency.CheckpointManager.'
    )

  def metadata(self, step: int | None = None) -> RootMetadata | StepMetadata:
    """Returns CheckpointManager level metadata if present, empty otherwise."""
    raise NotImplementedError(
        'Metadata not yet implemented for emergency.CheckpointManager.'
    )

  def metrics(self, step: int) -> Optional[PyTree]:
    """Returns metrics for step, if present."""
    raise NotImplementedError(
        'Metrics not yet implemented for emergency.CheckpointManager.'
    )

  def wait_until_finished(self):
    """Blocks until any incomplete save operations are completed.

    Note that this method will typically be a no-op if all checkpointers are
    synchronous, since old checkpoints are already cleaned up immediately after
    completing `save`, and there is no background thread to wait for.

    If some checkpointers are of type AsyncCheckpointer, however, this method
    will wait until each of these checkpointers is finished.
    """
    logging.info('Waiting for checkpoint to complete.')
    if self.in_primary_slice:
      self._persistent_checkpoint_manager.wait_until_finished()
    else:
      self._local_checkpoint_manager.wait_until_finished()

  def check_for_errors(self):
    """Checks for any outstanding errors in completed asynchronous save operations.

    Delegates to underlying Checkpointer.
    """
    if self.in_primary_slice:
      self._persistent_checkpoint_manager.check_for_errors()
    else:
      self._local_checkpoint_manager.check_for_errors()

  def close(self):
    """Waits for outstanding operations to finish and closes Checkpointers."""
    logging.info('Closing CheckpointManager.')
    self.wait_until_finished()
    if self.in_primary_slice:
      self._persistent_checkpoint_manager.close()
    else:
      self._local_checkpoint_manager.close()

  def __contextmanager__(
      self,
  ) -> Iterable[Self]:
    try:
      yield self
    finally:
      self.close()


class CheckpointManager(
    abstract_checkpoint_manager.AbstractCheckpointManager, epy.ContextManager
):
  """Provides both checkpoint management and emergency checkpointings.

  This class composes a local and a persistent checkpoint managers. The local
  manager saves checkpoints frequently to a fast local storage (like RAMFS).
  When a complete checkpoint exists at least one slice, restoration is possible,
  and the slice broadcasts the checkpoint to others. Additionally, the
  persistent manager checkpoints less frequently to a remote file system (e.g.,
  GCS),
  providing a fail-safe if local checkpoints become unavailable due to issues
  like hardware failure or preemption.

  Usage::

    options = CheckpointManagerOptions(
        local=LocalCheckpointOptions(save_interval_steps=2, max_to_keep=2),
        persistent=PersistentCheckpointOptions(
            save_interval_steps=5, max_to_keep=3
        ),
        enable_async_checkpointing=use_async,
    )
    return CheckpointManager(
        local_directory=local_directory,
        persistent_directory=persistent_directory,
        global_mesh=global_mesh,
        abstract_state=abstract_state,
        options=options,
    )
  """

  # TODO: b/330585086 - Support arbitrary items beyond state. We will have
  # to evaluate whether arbitrary items can be a good fit for local
  # checkpointing, given restore+broadcast requirements.
  def __init__(
      self,
      local_directory: epath.PathLike,
      persistent_directory: epath.PathLike,
      global_mesh: jax.sharding.Mesh,
      abstract_state: PyTree,  # a single PyTree describing the state structure
      *,
      options: Optional[CheckpointManagerOptions] = None,
      metadata: Optional[dict[str, Any]] = None,
      logger: Optional[abstract_logger.AbstractLogger] = None,
  ):
    options = options or CheckpointManagerOptions()
    self._global_mesh = global_mesh
    self._abstract_state = abstract_state
    self._slice_count = multislice.slice_count(
        global_mesh, replica_axis_index=options.replica_axis_index
    )
    checkpoint_manager._create_root_directory(
        persistent_directory,
        multiprocessing_options=checkpoint_manager.MultiprocessingOptions(),
    )
    if self._slice_count <= 0:
      raise ValueError(
          'Slice count must be positive, but got'
          f' {self._slice_count} for mesh {global_mesh}.'
      )
    elif self._slice_count == 1:
      del local_directory
      self._checkpoint_manager = checkpoint_manager.CheckpointManager(
          persistent_directory,
          options=_get_persistent_options(
              options, checkpoint_manager.MultiprocessingOptions()
          ),
          metadata=metadata,
          logger=logger,
      )
    else:
      self._checkpoint_manager = _MultisliceCheckpointManager(
          local_directory=local_directory,
          persistent_directory=persistent_directory,
          global_mesh=global_mesh,
          abstract_state=abstract_state,
          options=options,
          metadata=metadata,
          logger=logger,
      )

  @property
  def directory(self) -> epath.Path:
    return self._checkpoint_manager.directory

  @property
  def in_primary_slice(self) -> bool:
    if self._slice_count == 1:
      return True
    else:
      assert isinstance(self._checkpoint_manager, _MultisliceCheckpointManager)
      return self._checkpoint_manager.in_primary_slice

  @property
  def global_mesh(self) -> jax.sharding.Mesh:
    return self._global_mesh

  def all_steps(self, read: bool = False) -> Sequence[int]:
    return self._checkpoint_manager.all_steps(read=read)

  def latest_step(self) -> Optional[int]:
    return self._checkpoint_manager.latest_step()

  def best_step(self) -> Optional[int]:
    raise NotImplementedError(
        'Metrics tracking not yet implemented for emergency.CheckpointManager.'
    )

  def reload(self):
    return self._checkpoint_manager.reload()

  def reached_preemption(self, step: int) -> bool:
    return self._checkpoint_manager.reached_preemption(step)

  def should_save(self, step: int) -> bool:
    return self._checkpoint_manager.should_save(step)

  def delete(self, step: int):
    """Deletes a step checkpoint."""
    raise NotImplementedError(
        'Delete not yet implemented for emergency.CheckpointManager.'
    )

  def save(
      self,
      step: int,
      args: args_lib.Composite,
      *,
      force: bool = False,
  ) -> bool:
    return self._checkpoint_manager.save(step, args=args, force=force)

  def restore(
      self,
      step: int | None,
      args: args_lib.Composite | None = None,
  ) -> Any:
    del args
    args = args_lib.PyTreeRestore(
        item=self._abstract_state,
        restore_args=checkpoint_utils.construct_restore_args(
            self._abstract_state
        ),
    )
    return self._checkpoint_manager.restore(step, args=args)

  def item_metadata(self, step: int) -> Any:
    raise NotImplementedError(
        'Item metadata not implemented for emergency.CheckpointManager.'
    )

  def metadata(self, step: int | None = None) -> RootMetadata | StepMetadata:
    """Returns CheckpointManager level metadata if present, empty otherwise."""
    raise NotImplementedError(
        'Metadata not yet implemented for emergency.CheckpointManager.'
    )

  def metrics(self, step: int) -> Optional[PyTree]:
    """Returns metrics for step, if present."""
    raise NotImplementedError(
        'Metrics not yet implemented for emergency.CheckpointManager.'
    )

  def wait_until_finished(self):
    return self._checkpoint_manager.wait_until_finished()

  def check_for_errors(self):
    return self._checkpoint_manager.check_for_errors()

  def close(self):
    return self._checkpoint_manager.close()

  def __contextmanager__(
      self,
  ) -> Iterable[Self]:
    try:
      yield self
    finally:
      self.close()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of `CheckpointableHandler` for PyTrees."""

from __future__ import annotations

import asyncio
import json
from typing import Any, Awaitable

from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types

CheckpointableHandler = handler_types.CheckpointableHandler
JsonType = tree_types.JsonType


_DATA_FILENAME = 'data.json'


class JsonHandler(CheckpointableHandler[JsonType, None]):
  """An implementation of `CheckpointableHandler` for Json."""

  def __init__(self, filename: str | None = None):
    self._filename = filename or _DATA_FILENAME
    self._supported_filenames = [self._filename, _DATA_FILENAME, 'metadata']

  async def _background_save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: JsonType,
      *,
      primary_host: int | None = None,
  ):
    directory = await directory.await_creation()
    if multihost.is_primary_host(primary_host):
      path = directory / self._filename
      await asyncio.to_thread(path.write_text, json.dumps(checkpointable))

  async def save(
      self, directory: path_types.PathAwaitingCreation, checkpointable: JsonType
  ) -> Awaitable[None]:
    context = context_lib.get_context()
    return self._background_save(
        directory=directory,
        checkpointable=checkpointable,
        primary_host=context.multiprocessing_options.primary_host,
    )

  async def _background_load(
      self,
      directory: path_types.Path,
  ):
    for filename in self._supported_filenames:
      path = directory / filename
      if await asyncio.to_thread(path.exists):
        return json.loads(await asyncio.to_thread(path.read_text))
    raise FileNotFoundError(
        f'Unable to parse JSON file in {directory}. Recognized filenames are:'
        f' {self._supported_filenames}'
    )

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: None = None,
  ) -> Awaitable[JsonType]:
    return self._background_load(directory)

  async def metadata(self, directory: path_types.Path) -> None:
    return None

  def is_handleable(self, checkpointable: Any) -> bool:
    try:
      json.loads(json.dumps(checkpointable))
      return True
    except Exception:  # pylint: disable=broad-exception-caught
      return False

  def is_abstract_handleable(self, abstract_checkpointable: Any) -> bool | None:
    return None


class MetricsHandler(JsonHandler):

  def __init__(self):
    super().__init__(filename='metrics')



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry for `CheckpointableHandler`.

A `CheckpointableHandler` defines logic needed to save and restore a
"checkpointable" object. Once defined, the handler must be registered either
globally or locally.

To register the handler globally, use `register_handler`. Alternatively,
create a local registry confined to a specific scope by using
`local_registry` (note that globally-registered handlers are included in this
registry by default).

For example::

  ocp.handlers.register_handler(FooHandler)

  registry = ocp.handlers.local_registry()
  # `registry` already contains FooHandler.
  registry.add(BarHandler)
  # Scope this handler specifically to checkpointables named 'baz'.
  registry.add(BazHandler, 'baz')

  checkpointables_options = ocp.options.CheckpointablesOptions(
      registry=registry
  )
  with ocp.Context(checkpointables_options=checkpointables_options):
    ocp.save_checkpointables(...)

If a registered handler is scoped to a specific name (e.g.
`registry.add(BazHandler, 'baz')`), then this handler will always be
prioritized for saving or loading the checkpointable with that name, even if
the handler is not capable of saving/loading the checkpointable.

In the most common case, where a handler is not scoped to a specific name,
a given checkpointable (or abstract_checkpointable) will be resolved to a
handler returning True for `is_handleable` (or `is_abstract_handleable`),
respectively. If multiple handlers are usable, the first usable handler will be
returned. When loading, the handler type used for saving will be recorded in
the metadata, and will be used to resolve the handler, if a corresponding
handler is present in the registry.
"""

from __future__ import annotations

from typing import Any, Callable, Protocol, Sequence, Type, TypeVar

from absl import logging
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types

CheckpointableHandler = handler_types.CheckpointableHandler
RegistryEntry = tuple[Type[CheckpointableHandler], str | None]


def add_all(
    registry: CheckpointableHandlerRegistry,
    other_registry: CheckpointableHandlerRegistry,
) -> CheckpointableHandlerRegistry:
  """Adds all entries from `other_registry` to `registry`."""
  for handler, checkpointable in other_registry.get_all_entries():
    registry.add(handler, checkpointable)
  return registry


class CheckpointableHandlerRegistry(Protocol):
  """A registry defining a mapping from name to CheckpointableHandler.

  See module docstring for usage details.
  """

  def add(
      self,
      handler_type: Type[CheckpointableHandler],
      checkpointable: str | None = None,
  ) -> CheckpointableHandlerRegistry:
    """Adds an entry to the registry."""
    ...

  def get(
      self,
      checkpointable: str,
  ) -> Type[CheckpointableHandler]:
    """Gets the type of a `CheckpointableHandler` from the registry."""
    ...

  def has(
      self,
      checkpointable: str,
  ) -> bool:
    """Checks if an entry exists in the registry."""
    ...

  def get_all_entries(
      self,
  ) -> Sequence[RegistryEntry]:
    ...


class AlreadyExistsError(ValueError):
  """Raised when an entry already exists in the registry."""


class NoEntryError(KeyError):
  """Raised when no entry exists in the registry."""


class _DefaultCheckpointableHandlerRegistry(CheckpointableHandlerRegistry):
  """Default implementation of `CheckpointableHandlerRegistry`."""

  def __init__(
      self, other_registry: CheckpointableHandlerRegistry | None = None
  ):
    self._registry: list[RegistryEntry] = []

    # Initialize the registry with entries from other registry.
    if other_registry:
      add_all(self, other_registry)

  def add(
      self,
      handler_type: Type[CheckpointableHandler],
      checkpointable: str | None = None,
  ) -> CheckpointableHandlerRegistry:
    """Adds an entry to the registry.

    Args:
      handler_type: The handler type.
      checkpointable: The checkpointable name. If not-None, the registered
        handler will be scoped to that specific name. Otherwise, the handler
        will be available for any checkpointable name.

    Returns:
      The registry itself.

    Raises:
      AlreadyExistsError: If an entry for the given checkpointable name or
        handler type already exists in the registry.
      ValueError: If the handler is not default-constructible.
    """
    if not isinstance(handler_type, type):
      raise ValueError(
          f'The `handler_type` must be a type, but got {handler_type}.'
      )
    registered_handler_types = [
        handler_type for handler_type, _ in self.get_all_entries()
    ]
    if checkpointable:
      if self.has(checkpointable):
        raise AlreadyExistsError(
            f'Entry for checkpointable={checkpointable} already'
            ' exists in the registry.'
        )
    elif handler_type in registered_handler_types:
      raise AlreadyExistsError(
          f'Handler type {handler_type} already exists in the registry.'
      )
    self._registry.append((handler_type, checkpointable))
    return self

  def get(
      self,
      checkpointable: str,
  ) -> Type[CheckpointableHandler]:
    """Returns the handler for the given checkpointable name.

    Args:
      checkpointable: The checkpointable name.

    Returns:
      The handler for the given checkpointable name.

    Raises:
      NoEntryError: If no entry for the given checkpointable name exists in the
      registry.
    """
    for handler, checkpointable_name in self._registry:
      if checkpointable == checkpointable_name:
        return handler

    raise NoEntryError(
        f'No entry for checkpointable={checkpointable} in the registry.'
    )

  def has(
      self,
      checkpointable: str,
  ) -> bool:
    """Returns whether an entry for the given checkpointable name exists.

    Args:
      checkpointable: A checkpointable name.

    Returns:
      True if an entry for the given checkpointable name exists, False
      otherwise.
    """
    return any(
        checkpointable_name == checkpointable
        for _, checkpointable_name in self._registry
    )

  def get_all_entries(
      self,
  ) -> Sequence[RegistryEntry]:
    """Returns all entries in the registry."""
    return self._registry

  def __repr__(self):
    return f'_DefaultCheckpointableHandlerRegistry({self.get_all_entries()})'

  def __str__(self):
    return f'_DefaultCheckpointableHandlerRegistry({self.get_all_entries()})'


class ReadOnlyCheckpointableHandlerRegistry(CheckpointableHandlerRegistry):
  """Read-only implementation of `CheckpointableHandlerRegistry`."""

  def __init__(self, registry: CheckpointableHandlerRegistry):
    self._registry = registry

  def add(
      self,
      handler_type: Type[CheckpointableHandler],
      checkpointable: str | None = None,
  ) -> CheckpointableHandlerRegistry:
    raise NotImplementedError('Adding not implemented for read-only registry.')

  def get(
      self,
      checkpointable: str,
  ) -> Type[CheckpointableHandler]:
    return self._registry.get(checkpointable)

  def has(
      self,
      checkpointable: str,
  ) -> bool:
    return self._registry.has(checkpointable)

  def get_all_entries(
      self,
  ) -> Sequence[RegistryEntry]:
    return self._registry.get_all_entries()

  def __repr__(self):
    return f'ReadOnlyCheckpointableHandlerRegistry({self.get_all_entries()})'

  def __str__(self):
    return f'ReadOnlyCheckpointableHandlerRegistry({self.get_all_entries()})'


_GLOBAL_REGISTRY = _DefaultCheckpointableHandlerRegistry()


def global_registry() -> CheckpointableHandlerRegistry:
  """Returns the global registry."""
  return _GLOBAL_REGISTRY


def local_registry(
    other_registry: CheckpointableHandlerRegistry | None = None,
    *,
    include_global_registry: bool = True,
) -> CheckpointableHandlerRegistry:
  """Creates a local registry.

  Args:
    other_registry: An optional registry of handlers to include in the returned
      registry.
    include_global_registry: If true, includes globally-registered handlers in
      the returned registry by default.

  Returns:
    A local registry.
  """
  registry = _DefaultCheckpointableHandlerRegistry()
  if include_global_registry:
    registry = add_all(registry, global_registry())
  if other_registry:
    registry = add_all(registry, other_registry)
  return registry


CheckpointableHandlerType = TypeVar(
    'CheckpointableHandlerType', bound=Type[CheckpointableHandler]
)


def register_handler(
    cls: CheckpointableHandlerType,
) -> CheckpointableHandlerType:
  """Registers a :py:class:`.CheckpointableHandler` globally.

  The order in which handlers are registered matters. If multiple handlers
  could potentially be used to save or load, the one added most recently will be
  used.

  Usage::

    @ocp.handlers.register_handler
    class FooHandler(ocp.handlers.CheckpointableHandler[Foo, AbstractFoo]):
      ...

    ### OR ###

    ocp.handlers.register_handler(FooHandler)

  Args:
    cls: The handler class.

  Returns:
    The handler class.
  """
  _GLOBAL_REGISTRY.add(cls)
  return cls


def _construct_handler_instance(
    name: str,
    handler_type: Type[CheckpointableHandler],
) -> CheckpointableHandler:
  """Attempts to default-construct a handler type if possible."""
  assert isinstance(handler_type, type)
  try:
    return handler_type()
  except TypeError as e:
    raise ValueError(
        'The `CheckpointableHandler` resolved for'
        f' checkpointable={name} could not be'
        ' default-constructed. Please ensure the object is'
        ' default-constructible or provide a concrete instance.'
    ) from e


def _get_possible_handlers(
    registry: CheckpointableHandlerRegistry,
    is_handleable_fn: Callable[[CheckpointableHandler, Any], bool],
    checkpointable: Any | None,
    name: str,
) -> Sequence[CheckpointableHandler]:
  """Raises a NoEntryError if no possible handlers are found."""
  registry_entries = [
      (
          _construct_handler_instance(checkpointable_name, handler),
          checkpointable_name,
      )
      for handler, checkpointable_name in registry.get_all_entries()
  ]
  if checkpointable is None:
    # All handlers are potentially usable if checkpointable is not provided.
    possible_handlers = [
        handler
        for handler, checkpointable_name in registry_entries
        if checkpointable_name is None
    ]
  else:
    possible_handlers = [
        handler
        for handler, checkpointable_name in registry_entries
        if checkpointable_name is None
        and is_handleable_fn(handler, checkpointable)
    ]
  if not possible_handlers:
    available_handlers = [
        handler_type for handler_type, _ in registry.get_all_entries()
    ]
    error_msg = (
        f'Could not identify a valid handler for the checkpointable: "{name}"'
        f' and checkpointable type={type(checkpointable)}. Make sure to'
        ' register a `CheckpointableHandler` for the object using'
        ' `register_handler`, or by specifying a local registry'
        ' (`CheckpointablesOptions`). If a handler is already registered,'
        ' ensure that `is_handleable` correctly identifies the object as'
        f' handleable. The available handlers are: {available_handlers}'
    )
    raise NoEntryError(error_msg)
  return possible_handlers


def resolve_handler_for_save(
    registry: CheckpointableHandlerRegistry,
    checkpointable: Any,
    *,
    name: str,
) -> CheckpointableHandler:
  """Resolves a CheckpointableHandler for saving.

    1. If a name matching the provided checkpointable name is explicitly
       registered, return the corresponding handler.
    2. Resolve based on the `checkpointable` (using
      `CheckpointableHandler.is_handleable`).
    3. If multiple handlers are usable, return the *last* usable handler. This
       allows us to resolve the most recently-registered handler.

  Args:
    registry: The CheckpointableHandlerRegistry to search.
    checkpointable: A checkpointable to resolve.
    name: The name of the checkpointable.

  Raises:
    NoEntryError: If no compatible `CheckpointableHandler` can be found.

  Returns:
    A CheckpointableHandler instance.
  """
  # If explicitly registered, use that first.
  if registry.has(name):
    return _construct_handler_instance(name, registry.get(name))

  if checkpointable is None:
    raise ValueError('checkpointable must not be None for saving.')

  def is_handleable_fn(handler: CheckpointableHandler, ckpt: Any) -> bool:
    return handler.is_handleable(ckpt)

  possible_handlers = _get_possible_handlers(
      registry, is_handleable_fn, checkpointable, name
  )

  # Prefer the first handler in the absence of any other information.
  return possible_handlers[-1]


def resolve_handler_for_load(
    registry: CheckpointableHandlerRegistry,
    abstract_checkpointable: Any | None,
    *,
    name: str,
    handler_typestr: str,
) -> CheckpointableHandler:
  """Resolves a CheckpointableHandler for loading.

    1. If name is explicitly registered, return the handler.
    2. Resolve based on the `abstract_checkpointable` (using
      `CheckpointableHandler.is_abstract_handleable`).
    3. If `abstract_checkpointable` is None or not provided, all registered
      handlers not scoped to a specific item name are potentially usable.
    4. If multiple handlers are usable, return the handler with the matching
      typestr. If no matching typestr is found, then the handler used for saving
      may not be available now.
    4. Return the *last* usable handler. This allows us to resolve the most
       recently-registered handler.

  Raises:
    NoEntryError: If no compatible `CheckpointableHandler` can be found.

  Args:
    registry: The CheckpointableHandlerRegistry to search.
    abstract_checkpointable: An abstract checkpointable to resolve.
    name: The name of the checkpointable.
    handler_typestr: A CheckpointableHandler typestr to guide resolution.

  Returns:
    A CheckpointableHandler instance.
  """
  # If explicitly registered, use that first.
  if registry.has(name):
    return _construct_handler_instance(name, registry.get(name))

  def is_handleable_fn(
      handler: CheckpointableHandler, ckpt: Any
  ) -> bool | None:
    return handler.is_abstract_handleable(ckpt)

  possible_handlers = _get_possible_handlers(
      registry, is_handleable_fn, abstract_checkpointable, name
  )
  possible_handler_typestrs = [
      handler_types.typestr(type(handler)) for handler in possible_handlers
  ]

  try:
    idx = possible_handler_typestrs.index(handler_typestr)
    return possible_handlers[idx]
  except ValueError:
    logging.warning(
        'No handler found for typestr %s. The checkpointable may be restored'
        ' with different handler logic than was used for saving.',
        handler_typestr,
    )

  # Prefer the first handler in the absence of any other information.
  return possible_handlers[-1]



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ProtoHandler class."""

import asyncio
import inspect
from typing import Any, Awaitable, Type

from google.protobuf import message
from google.protobuf import text_format
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost


_DEFAULT_FILENAME = "proto.pbtxt"


class ProtoHandler(
    handler_types.CheckpointableHandler[message.Message, Type[message.Message]]
):
  """Serializes/deserializes protocol buffers."""

  def __init__(
      self,
      filename: str = _DEFAULT_FILENAME,
  ):
    """Initializes ProtoCheckpointHandler."""
    self._filename = filename

  async def _background_save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: message.Message,
      *,
      primary_host: int | None = None
  ):
    if multihost.is_primary_host(primary_host):
      directory = await directory.await_creation()
      path = directory / self._filename
      str_msg = text_format.MessageToString(checkpointable)
      await asyncio.to_thread(path.write_text, str_msg)

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: message.Message,
  ) -> Awaitable[None]:
    context = context_lib.get_context()
    return self._background_save(
        directory,
        checkpointable,
        primary_host=context.multiprocessing_options.primary_host,
    )

  async def _background_load(
      self,
      directory: path_types.Path,
      message_type: Type[message.Message],
  ):
    path = directory / self._filename
    str_msg = await asyncio.to_thread(path.read_text)
    return text_format.Parse(str_msg, message_type())

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: Type[message.Message] | None = None,
  ) -> Awaitable[message.Message]:
    if abstract_checkpointable is None:
      raise ValueError(
          "abstract_checkpointable must be provided to restore as a Proto."
      )
    return self._background_load(directory, abstract_checkpointable)

  async def metadata(self, directory: path_types.Path) -> Type[message.Message]:
    raise NotImplementedError()

  def is_handleable(self, checkpointable: Any) -> bool:
    return isinstance(checkpointable, message.Message)

  def is_abstract_handleable(self, abstract_checkpointable: Any) -> bool:
    return inspect.isclass(abstract_checkpointable) and issubclass(
        abstract_checkpointable, message.Message
    )
